import os
import typing as ty

import pandas as pd
import toml
from ETS_CookBook import ETS_CookBook as cook


# @cook.function_timer
def make_variants(case_name: str, csv_version: bool = True) -> None:

    if csv_version:
        make_csv_variants(case_name)
    else:
        make_toml_variants(case_name)


def make_csv_variants(case_name: str) -> None:
    '''
    This uses a csv, with commas as separators.
    If you want to put a list as a value, do it as such:
    "[1, 2.5, 5]"
    (with the quotes: note that this is not necessary in Excel, where you can
    just put the list with brackets without the quotes)
    '''

    variant_files_folder: str = f'variants/{case_name}/'
    variant_files: ty.List[str] = os.listdir(variant_files_folder)
    for variant_file in variant_files:

        reference_scenario_name: str = variant_file.split('.')[0]

        # Here we use a dict for the reference scenario
        # instead of a Box because we want to use functions
        # that use dicts to write a new scenario
        reference_scenario: ty.Dict = cook.parameters_from_TOML(
            f'scenarios/{case_name}/{reference_scenario_name}.toml'
        )

        variant_data: pd.DataFrame = pd.read_csv(
            f'{variant_files_folder}/{variant_file}'
        ).set_index('Variant name')

        variant_names: ty.List[str] = list(variant_data.index)

        for variant_name in variant_names:

            variant_scenario: ty.Dict = reference_scenario.copy()

            for quantity_string in variant_data.columns:

                variant_name_list = quantity_string.split('.')
                variant_value = variant_data.loc[variant_name][quantity_string]

                if isinstance(variant_value, float):
                    variant_value = float(variant_value)
                elif variant_value.startswith('['):
                    variant_value = variant_value.replace('[', '')
                    variant_value = variant_value.replace(']', '')
                    variant_value = variant_value.split(',')
                    variant_value = [
                        float(value_item) for value_item in variant_value
                    ]

                cook.set_nested_value(
                    variant_scenario, variant_name_list, variant_value
                )

            with open(
                f'scenarios/{case_name}/{variant_name}.toml',
                'w',
                encoding='utf-8',
            ) as file_to_write:
                toml.dump(variant_scenario, file_to_write)


def make_toml_variants(case_name: str) -> None:
    variant_configuration: ty.Dict = cook.parameters_from_TOML(
        f'variants/{case_name}.toml'
    )

    for reference_scenario_name in variant_configuration[
        'reference_scenarios'
    ]:
        # Here we use a dict for the reference scenario
        # instead of a Box because we want to use functions
        # that use dicts to write a new scenario
        reference_scenario: ty.Dict = cook.parameters_from_TOML(
            f'scenarios/{case_name}/{reference_scenario_name}.toml'
        )

        variant_names: ty.List[str] = variant_configuration[
            'reference_scenarios'
        ][reference_scenario_name]['variants']

        variant_scenarios: ty.Dict[str, ty.Dict] = {}
        for variant_name in variant_names:
            variant_scenarios[variant_name] = reference_scenario.copy()

        modified_parameters: ty.Dict = variant_configuration[
            'reference_scenarios'
        ][reference_scenario_name]['modified_parameters']

        modified_parameters_names: ty.List[ty.List[str]] = []

        for modified_parameter in modified_parameters:
            this_modification_names: ty.List[str] = [modified_parameter]
            next_step_in_modification = modified_parameters[modified_parameter]
            next_level: str = list(next_step_in_modification.keys())[0]
            while next_level not in variant_names:

                this_modification_names.append(next_level)
                next_step_in_modification = next_step_in_modification[
                    next_level
                ]
                next_level = list(next_step_in_modification.keys())[0]

            modified_parameters_names.append(this_modification_names)

        for modified_parameter_names in modified_parameters_names:
            for variant_name in variant_names:
                variant_scenario = variant_scenarios[variant_name]

                variant_nested_keys: ty.List[str] = (
                    modified_parameter_names + [variant_name]
                )
                modified_value = cook.get_nested_value(
                    modified_parameters, variant_nested_keys
                )
                cook.set_nested_value(
                    variant_scenario, modified_parameter_names, modified_value
                )

        for variant_name in variant_names:
            with open(
                f'scenarios/{case_name}/{variant_name}.toml',
                'w',
                encoding='utf-8',
            ) as file_to_write:
                toml.dump(variant_scenarios[variant_name], file_to_write)


if __name__ == '__main__':
    case_name = 'Mopo'
    csv_version = True
    make_variants(case_name, csv_version)



import os
import typing as ty

import pandas as pd
from ETS_CookBook import ETS_CookBook as cook


def test_battery_space() -> None:

    case_name: str = 'Mopo'
    general_parameters_file: str = 'ChaProEV.toml'
    general_parameters: ty.Dict = cook.parameters_from_TOML(
        general_parameters_file
    )
    save_interim_files: bool = general_parameters['interim_files']['pickle']

    if save_interim_files:
        scenario_files: ty.List[str] = os.listdir(f'scenarios/{case_name}')
        scenario_names: ty.List[str] = []

        for scenario_file in scenario_files:
            file_extension: str = scenario_file.split('.')[1]
            if file_extension == 'toml':
                scenario_names.append(scenario_file.split('.')[0])
        # scenario_names = [
        #         'XX_truck',
        #         'XX_van',
        #         'XX_bus',
        #     'XX_car'
        # ]
        for scenario_name in scenario_names:
            location_split: pd.DataFrame = pd.read_pickle(
                f'output/{case_name}/{scenario_name}_location_split.pkl'
            )
            sum_of_battery_spaces: pd.DataFrame = pd.read_pickle(
                f'output/{case_name}/{scenario_name}_sum_of_battery_spaces.pkl'
            )

            print(location_split.iloc[0:24])
            print(sum_of_battery_spaces.iloc[0:24])
            print(location_split.iloc[96:120])
            print(sum_of_battery_spaces.iloc[96:120])
            print(location_split.iloc[1102:1106])
            print(sum_of_battery_spaces.iloc[1102:1106])
            # exit()

            pd.testing.assert_frame_equal(
                location_split, sum_of_battery_spaces
            )


if __name__ == '__main__':
    test_battery_space()

    # staysa at route start?
    # cut at last day end
    # Do case without spillover?
    # weekend trips should have no spillover>\???? They do, but can be removed
    # easily



import datetime
import typing as ty

import numpy as np
import pandas as pd


def test_charge_and_consumption() -> None:

    # The charge and consuption might differ due to
    # boundary effects (i.e. if the run starts/stops at moments
    # where the consumption has happened, but the charge might not
    # have had the chance to occur)
    case_name: str = 'Mopo'
    test_start_year: int = 2018
    test_start_day: int = 1
    test_start_month: int = 1
    test_start_hour: int = 5
    test_end_year: int = 2018
    test_end_month: int = 12
    test_end_day: int = 31
    test_end_hour: int = 4
    test_start_date: datetime.datetime = datetime.datetime(
        year=test_start_year,
        month=test_start_month,
        day=test_start_day,
        hour=test_start_hour,
    )
    test_end_date: datetime.datetime = datetime.datetime(
        year=test_end_year,
        month=test_end_month,
        day=test_end_day,
        hour=test_end_hour,
    )

    test_range: pd.DatetimeIndex = pd.date_range(
        start=test_start_date, end=test_end_date, freq='1h', inclusive='both'
    )
    test_scenarios: ty.List[str] = [
        'XX_bus',
        'XX_van',
        'XX_truck',
        'XX_car'
    ]
    for scenario in test_scenarios:
        charge_drawn_by_vehicles_dataframe: pd.DataFrame = pd.read_pickle(
            f'output/{case_name}/{scenario}_charge_drawn_by_vehicles_total.pkl'
        )

        charge_drawn_by_vehicles_dataframe_in_test = (
            charge_drawn_by_vehicles_dataframe.loc[test_range]
        )
        charge_drawn_in_test: float = (
            charge_drawn_by_vehicles_dataframe_in_test[
                'Total Charge Drawn (kW)'
            ].sum()
        )

        consumption_dataframe: pd.DataFrame = pd.read_pickle(
            f'output/{case_name}/{scenario}_consumption_table.pkl'
        )
        consumption_dataframe_in_test = consumption_dataframe.loc[test_range]
        consumption_in_test: float = consumption_dataframe_in_test[
            'electricity consumption kWh'
        ].sum()
        print(scenario, f'{charge_drawn_in_test=}', f'{consumption_in_test=}')
        np.testing.assert_almost_equal(
            charge_drawn_in_test, consumption_in_test
        )


if __name__ == '__main__':
    test_charge_and_consumption()



'''
Author: Omar Usmani (Omar.Usmani@TNO.nl)
This module contains functions related to weather data and factors.
The weather data is pulled from the CDS (Climate Data Store) ERA-5 weather
data from the Copernicus institute
(https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-land?tab=form).
This data contains many quantities (such as temperature, precipitation,
or solar radiation) at an hourly level (starting in 1950) for the whole
world, at a one-decimal resolution for latitudes and longitudes.
Note that the data sometimes has trailing digits, but the resolution
still seems to be to the first decimal. This is the reason why we round the
coordinate values in our processing functions.
It contains the following functions:
1. **download_cds_weather_quantity:**
Downloads CDS ERA-5 weather data for a given quantity in a given area.
2. **download_all_cds_weather_data:**
Downloads all the necessary CDS weather data.
3. **make_weather_dataframe:** This function makes a weather DataFrame
into one we can use by
removing empty data and processing data into forms useful for the model.
4. **write_weather_database:** This function writes the weather database.
5. **get_hourly_values:** This function takes a Dataframe for a
given weather quantity. If this is a cumulative quantity, it adds hourly
values to it.
6. **get_all_hourly_values:** This functions adds hourly values to
cumulative quantities in the weather database.
7. **get_EV_tool_data:**: This gets the temperature efficiency data from the
     EV tool made by geotab.
    https://www.geotab.com/CMS-GeneralFiles-production/NA/EV/EVTOOL.html
8. **temperature_efficiency_factor:**This function returns the temperature
efficiency factor that corrects the baseline vehicle efficiency.
9. **plot_temperature_efficiency:** Plots the temperature efficiency
correction factor (source data versus interpolation)of electric vehicles.
10. **get_scenario_location_weather_quantity:** Returns a chosen weather
    quantity for a given location and a given runtime.
11. **get_scenario_weather_data:** Fetches the weather data and efficiency
    factors and puts it into a table that is saved to files/databases.
12. **solar_efficiency_factor:*** This gives us the efficiency factor of solar
     panels (i.e. how much of the solar radiation is converted into
     electricity).
    THIS IS A PLACEHOLDER FUNCTION
13. **setup_weather:** This runs all the functions necessary to get the
    scenario weather factors for a given case.
14. **get_location_weather_quantity:** Returns the value of a chosen
    weather quantity for a given location and time tag.
'''

import datetime
import os
import sqlite3
import typing as ty

import bs4
import cdsapi
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests
from ETS_CookBook import ETS_CookBook as cook


def download_cds_weather_quantity(
    quantity: str,
    year: str,
    month: str,
    days: ty.List[str],
    hours: ty.List[str],
    area: ty.List[float],
    download_folder: str,
) -> None:
    '''
    Downloads CDS (Climate Data Store) ERA-5 weather data for a given quantity
    in a given area (given by a list with:
    [latitude_min, latitude_max, longitude_min, longitude_max]).
    The days, months, hours numbers need two digits (so a leading zero is
    needed if the number only has one digit).
    Because of restrictions on the amount of data that can be downloaded at
    once (number of time points), we restrict this function to a given month
    (but hours and day can be all the hours and days in that month).
    To download data for a larger time span (say a year), you need
    to iterate.
    You only need to use this function at the start of your project,
    or if you change the general area you of the data you want to download
    (say, if you add locations in places for which you don't have data yet),
    or if you want to gather other quantities, or look at other years.
    '''
    quantity_download: cdsapi.Client = cdsapi.Client()
    cook.check_if_folder_exists(download_folder)
    quantity_download.retrieve(
        'reanalysis-era5-land',
        {
            'variable': quantity,
            'year': year,
            'month': month,
            'day': days,
            'time': hours,
            'area': area,
            'format': 'grib',
        },
        f'{download_folder}/{quantity}_{year}_{month}.grib',
    )


def download_all_cds_weather_data(
    scenario: ty.Dict, general_parameters: ty.Dict
) -> None:
    '''
    Downloads all the necessary CDS ERA-5 weather data.
    You only need to use this function at the start of your project,
    or if you change the general area you of the data you want to download
    (say, if you add locations in places for which you don't have data yet),
    or if you want to gather other quantities, or look at other years.
    '''

    time_parameters: ty.Dict = general_parameters['time']
    HOURS_IN_A_DAY: int = time_parameters['HOURS_IN_A_DAY']
    MONTHS_IN_A_YEAR: int = time_parameters['MONTHS_IN_A_YEAR']
    MAX_DAYS_IN_A_MONTH: int = time_parameters['MAX_DAYS_IN_A_MONTH']
    source_data_parameters: ty.Dict = scenario['weather']['source_data']
    start_year: int = source_data_parameters['start_year']
    end_year: int = source_data_parameters['end_year']

    years: ty.List[str] = [
        str(year) for year in range(start_year, end_year + 1)
    ]
    months: ty.List[str] = [
        f'{month_number:02d}'
        for month_number in range(1, MONTHS_IN_A_YEAR + 1)
    ]
    days: ty.List[str] = [
        f'{day_number:02d}' for day_number in range(1, MAX_DAYS_IN_A_MONTH + 1)
    ]
    quantities: ty.List[str] = source_data_parameters['quantities']
    hours: ty.List[str] = [
        f'{hour_number:02d}:00' for hour_number in range(HOURS_IN_A_DAY)
    ]
    latitude_min: float = source_data_parameters['latitude_min']
    latitude_max: float = source_data_parameters['latitude_max']
    longitude_min: float = source_data_parameters['longitude_min']
    longitude_max: float = source_data_parameters['longitude_max']
    raw_data_folder: str = source_data_parameters['raw_data_folder']
    area: ty.List[float] = [
        latitude_min,
        longitude_min,
        latitude_max,
        longitude_max,
    ]

    for quantity in quantities:
        for year in years:
            for month in months:
                print(quantity, year, month)
                download_cds_weather_quantity(
                    quantity, year, month, days, hours, area, raw_data_folder
                )


def make_weather_dataframe(
    raw_weather_dataframe: pd.DataFrame,
    quantity: str,
    quantity_tag: str,
    quantity_name: str,
    scenario: ty.Dict,
) -> pd.DataFrame:
    '''
    This function makes a weather DataFrame into one we can use by
    processing data into forms useful for the model.
    The processed DataFrame can then be added to the weather database.
    This weather database has an area given by latitude and longitude
    ranges. Note that these ranges correspond to the general region of the
    scenario, and thus can be different from the download range
    (which can be for example the whole of Europe) and from the scenario range
    (which corresponds to all locations in the scenario).
    Typically, you will want to have a large range for the downloaded data
    (which you would do only once), a medium range for the processed weather
    data (which would cover the range of locations in the scenarios you are
    planning to do in the mid-term), and scenario-specific locations (with
    their own latitudes and longitudes).
    '''

    weather_parameters: ty.Dict = scenario['weather']['processed_data']
    latitude_min: float = weather_parameters['latitude_min']
    latitude_max: float = weather_parameters['latitude_max']
    longitude_min: float = weather_parameters['longitude_min']
    longitude_max: float = weather_parameters['longitude_max']
    coordinate_step: float = weather_parameters['coordinate_step']
    # We round the coordinates to the first decimal,
    # as this is the data resolution (but with trailing digits).
    # (See general module exaplanation).
    latitudes: ty.List[float] = [
        np.round(latitude, 1)
        for latitude in np.arange(
            latitude_min, latitude_max + coordinate_step, coordinate_step
        )
    ]
    longitudes: ty.List[float] = [
        np.round(longitude, 1)
        for longitude in np.arange(
            longitude_min, longitude_max + coordinate_step, coordinate_step
        )
    ]
    KELVIN_TO_CELSIUS: float = weather_parameters['KELVIN_TO_CELSIUS']

    temperature_quantities: ty.List[str] = weather_parameters[
        'temperature_quantities'
    ]
    processed_index_tags: ty.List[str] = weather_parameters[
        'processed_index_tags'
    ]

    # We slice the latitudes and longitudes
    # Note that we round the values to the first decimal,
    # as this is the data resolution (but with trailing digits).
    # (See general module exaplanation).
    raw_weather_dataframe = raw_weather_dataframe.loc[
        np.round(
            raw_weather_dataframe.index.get_level_values('latitude'), 1
        ).isin(latitudes)
    ]

    raw_weather_dataframe = raw_weather_dataframe.loc[
        np.round(
            raw_weather_dataframe.index.get_level_values('longitude'), 1
        ).isin(longitudes)
    ]

    # We process the time data into time tags (if needed)
    if 'valid_time' in raw_weather_dataframe.columns:
        weather_data_time_tags: pd.Index = pd.Index(
            raw_weather_dataframe['valid_time'].values
        )
    else:
        raw_times: pd.Index = raw_weather_dataframe.index.get_level_values(
            'time'
        )
        raw_steps: pd.Index = raw_weather_dataframe.index.get_level_values(
            'step'
        )
        weather_data_time_tags = raw_times + raw_steps

    # We round the latitudes and longitudes (see above) and gather the values
    raw_latitudes: pd.Index = raw_weather_dataframe.index.get_level_values(
        'latitude'
    )
    raw_longitudes: pd.Index = raw_weather_dataframe.index.get_level_values(
        'longitude'
    )
    latitudes = np.round(raw_latitudes, 1)
    longitudes = np.round(raw_longitudes, 1)
    values: pd.api.extensions.ExtensionArray | np.ndarray = (
        raw_weather_dataframe[quantity_tag].values
    )

    if quantity in temperature_quantities:
        kelvin_to_celsius_conversion: np.ndarray = KELVIN_TO_CELSIUS * np.ones(
            len(values)
        )
        values = values + kelvin_to_celsius_conversion

    # We can collect everyting into a processed dataframe
    processed_weather_dataframe: pd.DataFrame = pd.DataFrame(
        values,
        columns=[quantity_name],
        index=[weather_data_time_tags, latitudes, longitudes],
    )

    processed_weather_dataframe.index.rename(
        processed_index_tags, inplace=True
    )

    return processed_weather_dataframe


def write_weather_database(scenario: ty.Dict) -> None:
    '''
    This function writes the weather database.
    It iterates over desired quantities, processes
    the corresponding .grib source files into dataframes
    (including some processing), and writes them to a database.
    '''

    weather_parameters: ty.Dict = scenario['weather']['processed_data']
    raw_data_folder: str = weather_parameters['raw_data_folder']
    processed_folder: str = weather_parameters['processed_folder']
    weather_database_file_name: str = weather_parameters[
        'weather_database_file_name'
    ]
    quantities: ty.List[str] = weather_parameters['quantities']
    quantity_tags: ty.List[str] = weather_parameters['quantity_tags']
    quantity_processed_names: ty.List[str] = weather_parameters[
        'quantity_processed_names'
    ]
    chunk_size: int = weather_parameters['chunk_size']
    cook.check_if_folder_exists(processed_folder)
    weather_database_file: str = (
        f'{processed_folder}/{weather_database_file_name}'
    )

    for quantity, quantity_tag, quantity_name in zip(
        quantities, quantity_tags, quantity_processed_names
    ):
        # We create a new table for each quantity
        clear_table: bool = True
        print(quantity)
        for file_name in os.listdir(raw_data_folder):
            if file_name.split('.')[-1] == 'grib':
                file_header = file_name.split('.')[0]
                if file_header[0 : len(quantity)] == quantity:
                    source_file: str = f'{raw_data_folder}/{file_name}'
                    quantity_dataframe: pd.DataFrame = (
                        cook.from_grib_to_dataframe(source_file)
                    )
                    quantity_dataframe = quantity_dataframe.dropna(
                        subset=[quantity_tag]
                    )

                    processed_weather_dataframe: pd.DataFrame = (
                        make_weather_dataframe(
                            quantity_dataframe,
                            quantity,
                            quantity_tag,
                            quantity_name,
                            scenario,
                        )
                    )
                    cook.put_dataframe_in_sql_in_chunks(
                        processed_weather_dataframe,
                        weather_database_file,
                        quantity_name,
                        chunk_size,
                        drop_existing_table=clear_table,
                    )
                    # In following iteraions, we just want to
                    # append data to an existing table.
                    clear_table = False


def get_hourly_values(
    quantity_dataframe: pd.DataFrame,
    quantity: str,
    quantity_name: str,
    scenario: ty.Dict,
) -> pd.DataFrame:
    '''
    This function takes a Dataframe for a give weather quantity.
    If this is a cumulative quantity, it adds hourly values to it.
    '''
    weather_parameters: ty.Dict = scenario['weather']['processed_data']
    cumulative_quantities: ty.List[str] = weather_parameters[
        'cumulative_quantities'
    ]

    latitudes: np.ndarray = np.unique(
        quantity_dataframe.index.get_level_values('Latitude')
    )
    longitudes: np.ndarray = np.unique(
        quantity_dataframe.index.get_level_values('Longitude')
    )

    if quantity in cumulative_quantities:
        quantity_dataframe[f'Shifted {quantity_name}'] = [np.nan] * len(
            quantity_dataframe.index
        )
        quantity_dataframe[f'Hourly {quantity_name}'] = [np.nan] * len(
            quantity_dataframe.index
        )
        # We need to do this for each latitude/longitude combination
        # We will take a slice and fill it with values
        # This will also update the main dataframe

        for latitude in latitudes:
            for longitude in longitudes:
                # We need to check if the latitude/longitude combination
                # has data
                if (latitude, longitude) in quantity_dataframe.index:
                    location_dataframe = quantity_dataframe.loc[
                        latitude, longitude
                    ]

                    cumulative_values: ty.List[float] = location_dataframe[
                        quantity_name
                    ].values
                    shifted_cumulative_values: np.ndarray = np.roll(
                        cumulative_values, -1
                    )
                    location_dataframe.loc[:, f'Shifted {quantity_name}'] = (
                        shifted_cumulative_values
                    )

                    location_dataframe.loc[:, f'Hourly {quantity_name}'] = (
                        shifted_cumulative_values
                        - cumulative_values
                        * (
                            location_dataframe.index.get_level_values(
                                'Timetag'
                            ).hour
                            != 0
                        )
                        # At midnight, the shifted value is equal to the
                        # hourly value (it's the cumulative value
                        # of the time from 00.00  and 01.00).
                    )

                    # The data seems to have some issues
                    # where the cumulative values
                    # sometimes decrease (slightly), so we change these to zero
                    location_dataframe.loc[
                        location_dataframe[f'Hourly {quantity_name}'] < 0,
                        f'Hourly {quantity_name}',
                    ] = 0

                    # The last value needs to be set to zero,
                    # as it will be based on the first value
                    # (but needs to be based
                    # on the following year, which is not in the data)
                    location_dataframe.at[
                        location_dataframe.index.values[-1],
                        f'Hourly {quantity_name}',
                    ] = 0

    return quantity_dataframe


def get_all_hourly_values(scenario: ty.Dict) -> None:
    '''
    This functions adds hourly values to cumulative quantities
    in the weather database.
    '''
    weather_parameters: ty.Dict = scenario['weather']['processed_data']

    quantities: ty.List[str] = weather_parameters['quantities']
    cumulative_quantities: ty.List[str] = weather_parameters[
        'cumulative_quantities'
    ]
    processed_index_tags: ty.List[str] = weather_parameters[
        'processed_index_tags'
    ]
    processed_folder: str = weather_parameters['processed_folder']
    weather_database_file_name: str = weather_parameters[
        'weather_database_file_name'
    ]
    weather_database: str = f'{processed_folder}/{weather_database_file_name}'
    quantity_names: ty.List[str] = weather_parameters[
        'quantity_processed_names'
    ]
    chunk_size: int = weather_parameters['chunk_size']
    queries_for_cumulative_quantities: ty.List[str] = weather_parameters[
        'queries_for_cumulative_quantities'
    ]

    query_dictionary: ty.Dict[str, str] = dict(
        zip(cumulative_quantities, queries_for_cumulative_quantities)
    )

    for quantity, quantity_name in zip(quantities, quantity_names):
        if quantity in cumulative_quantities:
            with sqlite3.connect(weather_database) as sql_connection:
                sql_query: str = query_dictionary[quantity]
                quantity_dataframe: pd.DataFrame = pd.read_sql(
                    sql_query, sql_connection
                )
                quantity_dataframe['Timetag'] = pd.to_datetime(
                    quantity_dataframe['Timetag']
                )
                quantity_dataframe = quantity_dataframe.set_index(
                    processed_index_tags
                )

            print('Avoid swap by arranging before? And explain/improve')
            quantity_dataframe = quantity_dataframe.swaplevel(0, 2)
            quantity_dataframe = quantity_dataframe.swaplevel(0, 1)
            quantity_dataframe = quantity_dataframe.sort_index()

            quantity_dataframe = get_hourly_values(
                quantity_dataframe, quantity, quantity_name, scenario
            )

            cook.put_dataframe_in_sql_in_chunks(
                quantity_dataframe, weather_database, quantity_name, chunk_size
            )


def get_EV_tool_data(scenario: ty.Dict) -> None:
    '''
    This gets the temperature efficiency data from the EV tool made
    by geotab.
    https://www.geotab.com/CMS-GeneralFiles-production/NA/EV/EVTOOL.html
    '''

    EV_tool_parameters: ty.Dict = scenario['weather']['EV_tool']

    EV_tool_url: str = EV_tool_parameters['EV_tool_url']
    user_agent: str = EV_tool_parameters['user_agent']

    EV_tool_session: requests.Session = requests.Session()
    EV_tool_session.headers['User-Agent'] = user_agent
    EV_tool_html_content: bytes = EV_tool_session.get(EV_tool_url).content
    EV_tool_soup: bs4.BeautifulSoup = bs4.BeautifulSoup(
        EV_tool_html_content, 'html.parser'
    )

    EV_tool_scripts: ty.List[bs4.Script] = []
    for script in EV_tool_soup.find_all('script'):
        EV_tool_scripts.append(script)

    efficiency_curve_script_index: int = EV_tool_parameters[
        'efficiency_curve_script_index'
    ]
    efficiency_curve_script: bs4.Script = EV_tool_scripts[
        efficiency_curve_script_index
    ]
    efficiency_curve_script_text: str = efficiency_curve_script.get_text()
    data_splitter: str = EV_tool_parameters['data_splitter']
    efficiency_curve_script_data: ty.List[str] = (
        efficiency_curve_script_text.split(data_splitter)
    )

    temperatures: ty.List[float] = []
    efficiency_factors: ty.List[float] = []

    for data_point in efficiency_curve_script_data:
        data_values: ty.List[str] = data_point.split(',')
        if len(data_values) > 1:
            # Some entries are not actual data entries and are empty
            temperature: float = float(data_values[0].strip(" '"))
            temperatures.append(temperature)
            raw_efficiency_factor: str = data_values[1].split(':')[1]
            efficiency_factor = float(raw_efficiency_factor.split("'")[1])
            efficiency_factors.append(efficiency_factor)
    efficiency_factor_column_name: str = EV_tool_parameters[
        'efficiency_factor_column_name'
    ]
    temperature_efficiencies: pd.DataFrame = pd.DataFrame(
        efficiency_factors,
        columns=[efficiency_factor_column_name],
        index=temperatures,
    )

    temperature_efficiencies.index.name = 'Temperature (°C)'
    # This tag (the ° symbol, probably) creates issues with xml files
    # (and a warning with stata), so we need to remove the xml output
    # (or change things).
    file_name: str = EV_tool_parameters['file_name']
    groupfile_name: str = EV_tool_parameters['groupfile_name']
    folder: str = EV_tool_parameters['folder']
    cook.save_dataframe(
        temperature_efficiencies, file_name, groupfile_name, folder, scenario
    )


def temperature_efficiency_factor(
    temperature: float, scenario: ty.Dict
) -> float:
    '''
    This function returns the temperature efficiency factor that corrects
    the baseline vehicle efficiency. It uses a data file (extracted from
    here https://www.geotab.com/CMS-GeneralFiles-production/NA/EV/EVTOOL.html).
    It then uses a fitting function. This is done for the following reasons:
    1) We can go beyond the temperature range provided by the data
    2) A function works better than a data lookup for future uses
    3) Consistency between the temparatures in and out of data range.
    '''

    EV_tool_parameters: ty.Dict = scenario['weather']['EV_tool']

    vehicle_temperature_efficiencies_parameters: ty.Dict = scenario['weather'][
        'vehicle_temperature_efficiencies'
    ]

    source_folder: str = vehicle_temperature_efficiencies_parameters['folder']
    source_file: str = vehicle_temperature_efficiencies_parameters['file_name']
    values_header: str = EV_tool_parameters['efficiency_factor_column_name']
    fitting_polynomial_order: int = (
        vehicle_temperature_efficiencies_parameters['fitting_polynomial_order']
    )

    temperature_efficiencies_data: pd.DataFrame = pd.read_pickle(
        f'{source_folder}/{source_file}'
    )
    data_temperature_range: pd.Index[float] = (
        temperature_efficiencies_data.index
    )
    data_efficiencies: pd.Series[float] = temperature_efficiencies_data[
        values_header
    ]
    fitting_function_coefficients: np.ndarray = np.polyfit(
        data_temperature_range, data_efficiencies, fitting_polynomial_order
    )
    fitting_function: np.poly1d = np.poly1d(fitting_function_coefficients)

    efficiency_factor: float = fitting_function(temperature)

    return efficiency_factor


def plot_temperature_efficiency(scenario: ty.Dict) -> None:
    '''
    Plots the temperature efficiency correction factor
    (source data versus interpolation)
    of electric vehicles.
    '''

    EV_tool_parameters: ty.Dict = scenario['weather']['EV_tool']

    plot_colors: pd.DataFrame = cook.get_extra_colors(scenario)

    plot_parameters: ty.Dict = scenario['plots'][
        'vehicle_temperature_efficiency'
    ]
    plot_style: str = plot_parameters['style']
    geotab_data_color: str = plot_colors.loc[
        plot_parameters['geotab_data_color']
    ]
    geotab_data_size: int = plot_parameters['geotab_data_size']
    fit_color: str = plot_colors.loc[plot_parameters['fit_color']]
    fit_line_size: int = plot_parameters['fit_line_size']
    title_size: int = plot_parameters['title_size']
    plt.style.use(plot_style)
    source_data_folder: str = plot_parameters['source_data_folder']
    source_data_file: str = plot_parameters['source_data_file']
    source_data: pd.DataFrame = pd.read_pickle(
        f'{source_data_folder}/{source_data_file}'
    )
    temperatures: np.ndarray = source_data.index.values
    values_header: str = EV_tool_parameters['efficiency_factor_column_name']
    geotab_data_efficiencies: pd.api.extensions.ExtensionArray | np.ndarray = (
        source_data[values_header].values
    )
    fitted_efficiencies: ty.List[float] = [
        temperature_efficiency_factor(temperature, scenario)
        for temperature in temperatures
    ]

    temeprature_efficiency_figure, temperature_efficiency_plot = plt.subplots(
        1, 1
    )
    temperature_efficiency_plot.plot(
        temperatures,
        geotab_data_efficiencies,
        label='geotab_data',
        marker='.',
        markersize=geotab_data_size,
        color=geotab_data_color,
        linestyle='none',
    )
    temperature_efficiency_plot.plot(
        temperatures,
        fitted_efficiencies,
        label='fit',
        linewidth=fit_line_size,
        color=fit_color,
    )
    temperature_efficiency_plot.legend()
    temperature_efficiency_plot.set_xlabel('Temperature (°C)')
    temperature_efficiency_plot.set_ylabel('Efficiency factor')
    temperature_efficiency_plot.set_title(
        'Temperature correction factor of range/efficiency of BEVs',
        fontsize=title_size,
    )
    temperature_efficiency_plot.set_ylim(0, 1.05)
    temeprature_efficiency_figure.tight_layout()
    cook.save_figure(
        temeprature_efficiency_figure,
        'Vehicle_Temperature_correction_factor',
        source_data_folder,
        scenario,
    )


def get_scenario_location_weather_quantity(
    location_latitude: float,
    location_longitude: float,
    run_start: datetime.datetime,
    run_end: datetime.datetime,
    source_table: str,
    weather_quantity: str,
    scenario: ty.Dict,
) -> pd.DataFrame:
    '''
    Returns a chosen weather quantity for a given location and a given
    runtime.
    The run_start and run_end inputs are datetime objects,
    source_table and weather_quantity sometimes have the same name,
    but not always (e.g. for hourly values of a cumulative quantity
    such as the solar radiation downwards)
    '''

    processed_data_parameters: ty.Dict = scenario['weather']['processed_data']
    processed_folder: str = processed_data_parameters['processed_folder']
    weather_database_file_name: str = processed_data_parameters[
        'weather_database_file_name'
    ]
    weather_database_connection: sqlite3.Connection = sqlite3.connect(
        f'{processed_folder}/{weather_database_file_name}'
    )

    # We need to avoid issues with spaces in column names, so we need
    # nested double quotes
    source_table_text: str = f'"{source_table}"'
    weather_quantity_text: str = f'"{weather_quantity}"'
    run_start_text: str = f'"{run_start}"'
    run_end_text: str = f'"{run_end}"'

    list_columns_to_fetch: ty.List[str] = [
        'Latitude',
        'Longitude',
        'Timetag',
        weather_quantity_text,
    ]
    # We need to convert this into a string for a query
    columns_to_fetch: str = ','.join(list_columns_to_fetch)

    query_filter_quantities: ty.List[str] = [
        'Latitude',
        'Longitude',
        'Timetag',
    ]
    query_filter_types: ty.List[str] = ['=', '=', 'between']
    query_filter_values: ty.List = [
        str(location_latitude),
        str(location_longitude),
        [run_start_text, run_end_text],
    ]

    location_scenario_query: str = cook.read_query_generator(
        columns_to_fetch,
        source_table_text,
        query_filter_quantities,
        query_filter_types,
        query_filter_values,
    )

    weather_values: pd.DataFrame = pd.read_sql(
        location_scenario_query, weather_database_connection
    )

    # We don't need the latitude and longitude values in columns
    # (they are all the same and are our input)
    weather_values = weather_values.drop(columns=['Latitude', 'Longitude'])

    return weather_values


def get_scenario_weather_data(
    scenario: ty.Dict, case_name: str, general_parameters: ty.Dict
) -> pd.DataFrame:
    '''
    Fetches the weather data and efficiency factors and puts it into
    a table that is saved to files/databases.
    '''
    weather_dataframe: pd.DataFrame = pd.DataFrame()

    scenario_name: str = scenario['scenario_name']

    weather_factors_table_root_name: str = scenario['weather'][
        'weather_factors_table_root_name'
    ]
    weather_factors_table_name: str = (
        f'{scenario_name}_{weather_factors_table_root_name}'
    )

    file_parameters: ty.Dict = general_parameters['files']
    output_folder: str = f'{file_parameters["output_root"]}/{case_name}'
    groupfile_root: str = file_parameters['groupfile_root']
    groupfile_name: str = f'{groupfile_root}_{case_name}'

    weather_processed_data_parameters: ty.Dict = scenario['weather'][
        'processed_data'
    ]
    quantity_processed_names: ty.List[str] = weather_processed_data_parameters[
        'quantity_processed_names'
    ]

    cumulative_quantity_processed_names: ty.List[str] = (
        weather_processed_data_parameters[
            'cumulative_quantity_processed_names'
        ]
    )

    run_parameters: ty.Dict = scenario['run']
    run_start: datetime.datetime = datetime.datetime(
        run_parameters['start']['year'],
        run_parameters['start']['month'],
        run_parameters['start']['day'],
        run_parameters['start']['hour'],
        run_parameters['start']['minute'],
    )
    run_end: datetime.datetime = datetime.datetime(
        run_parameters['end']['year'],
        run_parameters['end']['month'],
        run_parameters['end']['day'],
        run_parameters['end']['hour'],
        run_parameters['end']['minute'],
    )

    location_parameters: ty.Dict = scenario['locations']

    for location in location_parameters:
        location_weather_dataframe: pd.DataFrame = pd.DataFrame()
        location_first_quantity: bool = True
        location_latitude: float = location_parameters[location]['latitude']
        location_longitude: float = location_parameters[location]['longitude']

        for quantity in quantity_processed_names:
            source_table: str = quantity
            if quantity in cumulative_quantity_processed_names:
                weather_quantity: str = f'Hourly {quantity}'
            else:
                weather_quantity = quantity

            weather_values: pd.DataFrame = (
                get_scenario_location_weather_quantity(
                    location_latitude,
                    location_longitude,
                    run_start,
                    run_end,
                    source_table,
                    weather_quantity,
                    scenario,
                )
            )
            if location_first_quantity:
                location_weather_dataframe['Location'] = [location] * len(
                    weather_values.index
                )
                location_weather_dataframe['Timetag'] = weather_values[
                    'Timetag'
                ]
                location_weather_dataframe = (
                    location_weather_dataframe.set_index(
                        ['Location', 'Timetag']
                    )
                )
                location_first_quantity = False

            location_weather_dataframe[weather_quantity] = weather_values[
                weather_quantity
            ].values

        weather_dataframe = pd.concat(
            [weather_dataframe, location_weather_dataframe], ignore_index=False
        )

        weather_dataframe['Vehicle efficiency factor'] = [
            temperature_efficiency_factor(
                temperature_to_use,
                scenario,
            )
            for temperature_to_use in weather_dataframe[
                'Temperature at 2 meters (°C)'
            ].values
        ]

        JOULES_IN_A_KWH: float = general_parameters['unit_conversions'][
            'JOULES_IN_A_KWH'
        ]
        weather_dataframe['Vehicle solar panels production (kWhe/m2)'] = [
            weather_dataframe[
                'Hourly Surface solar radiation downwards (J/m2)'
            ]
            * solar_panels_efficiency_factor(temperature_to_use)
            / JOULES_IN_A_KWH
            for temperature_to_use in weather_dataframe[
                'Temperature at 2 meters (°C)'
            ]
        ]

    cook.save_dataframe(
        weather_dataframe,
        weather_factors_table_name,
        groupfile_name,
        output_folder,
        scenario,
    )

    return weather_dataframe


def solar_panels_efficiency_factor(temperature: float) -> float:
    '''
    This gives us the efficiency factor of solar panels (i.e. how
    much of the solar radiation is converted into electricity).
    THIS IS A PLACEHOLDER FUNCTION
    '''
    # THIS IS A PLACEHOLDER VALUE
    efficiency_factor: float = np.exp(-((temperature - 25) ** 2) / (100)) / 3

    return efficiency_factor


def setup_weather(
    scenario: ty.Dict, case_name: str, general_parameters: ty.Dict
) -> None:
    '''
    This runs all the functions necessary to get the scenario weather factors
    for a given case. Downloading CDS weather data
    (which is in principle only done once, but can be repeated if we add
    new years or new areas) is an option. The same holds for the EV tool
    temperature curve data.
    Creating the weather database is also optional (it should happen less
    often than updates related to scenarios, but more often than the above
    downloads).
    '''

    get_extra_downloads: ty.Dict[str, bool] = scenario['run'][
        'get_extra_downloads'
    ]
    download_weather_data: bool = get_extra_downloads['download_weather_data']
    download_EV_tool_data: bool = get_extra_downloads['download_EV_tool_data']
    make_weather_database: bool = get_extra_downloads['make_weather_database']

    if download_weather_data:
        download_all_cds_weather_data(scenario, general_parameters)
    if download_EV_tool_data:
        get_EV_tool_data(scenario)
        plot_temperature_efficiency(scenario)
    if make_weather_database:
        try:
            write_weather_database(scenario)
        except FileNotFoundError:
            raise FileNotFoundError(
                'Data files for writing the weather database not found.'
                'You can download them by setting download_weather_data as'
                'true (lowercase is required by toml) under the'
                '[run.get_extra_downloads] in your scenario file.'
            )

        get_all_hourly_values(scenario)
    try:
        get_scenario_weather_data(scenario, case_name, general_parameters)
    except sqlite3.OperationalError:
        raise FileNotFoundError(
            'Weather database could not be opened, it may not exist.'
            'A possible solution is to create the weather database. '
            'To do so, set make_weather_database as true'
            '(lowercase is required by toml) under [run.get_extra_downloads]'
            'in your scenario file.'
        )
    except FileNotFoundError:
        raise FileNotFoundError(
            'Vehicle Efficiency data not found.'
            'A possible solution is to set download_EV_tool_data as true'
            '(lowercase is required by toml) under [run.get_extra_downloads]'
            'in your scenario file.'
        )


def get_location_weather_quantity(
    location_latitude: float,
    location_longitude: float,
    timetag: datetime.datetime,
    source_table: str,
    weather_quantity: str,
    scenario: ty.Dict,
) -> float:
    '''
    Returns the value a a chosen weather quantity for a given location and
    time tag.
    '''

    processed_data_parameters: ty.Dict = scenario['weather']['processed_data']
    processed_folder: str = processed_data_parameters['processed_folder']
    weather_database_file_name: str = processed_data_parameters[
        'weather_database_file_name'
    ]
    weather_database_connection: sqlite3.Connection = sqlite3.connect(
        f'{processed_folder}/{weather_database_file_name}'
    )
    # We need to avoid issues with spaces in column names, so we need
    # nested double quotes
    source_table = f'"{source_table}"'
    weather_quantity = f'"{weather_quantity}"'
    timetag_text: str = f'"{timetag}"'

    list_columns_to_fetch: ty.List[str] = [
        'Latitude',
        'Longitude',
        'Timetag',
        weather_quantity,
    ]
    # We need to convert this into a string for a query
    columns_to_fetch: str = ','.join(list_columns_to_fetch)

    query_filter_quantities: ty.List[str] = [
        'Latitude',
        'Longitude',
        'Timetag',
    ]
    query_filter_types: ty.List[str] = ['=', '=', '=']
    query_filter_values: ty.List = [
        location_latitude,
        location_longitude,
        timetag_text,
    ]

    scenario_query: str = cook.read_query_generator(
        columns_to_fetch,
        source_table,
        query_filter_quantities,
        query_filter_types,
        query_filter_values,
    )

    weather_values: pd.DataFrame = pd.read_sql(
        scenario_query, weather_database_connection
    )

    # The column does not contain double quotes and we need
    # the value (hence [0])
    value_of_weather_quantity: float = weather_values[
        weather_quantity.strip('"')
    ][0]

    return value_of_weather_quantity


if __name__ == '__main__':
    general_parameters_file_name: str = 'ChaProEV.toml'
    general_parameters: ty.Dict = cook.parameters_from_TOML(
        general_parameters_file_name
    )
    case_name = 'local_impact_BEVs'
    test_scenario_name: str = 'baseline'
    scenario_file_name: str = (
        f'scenarios/{case_name}/{test_scenario_name}.toml'
    )
    scenario: ty.Dict = cook.parameters_from_TOML(scenario_file_name)
    scenario['scenario_name'] = test_scenario_name
    plot_temperature_efficiency(scenario)
    setup_weather(scenario, case_name, general_parameters)
    print(
        get_location_weather_quantity(
            52.0,
            4.2,
            datetime.datetime(2020, 5, 8, 8, 0),
            'Temperature at 2 meters (°C)',
            'Temperature at 2 meters (°C)',
            scenario,
        )
    )



'''
Author: Omar Usmani (Omar.Usmani@TNO.nl)
This module defines and declares classes for the different objects
that define the system (the parameters/defintions come from a scenario file),
namely:
1. **Legs:** Legs are point-to-point vehicle movements (i.e. movements where
    the vehicle goes from a start location and ends/stops at an end location).
2. **Location:** This class defines the locations where the vehicles are
(available charger power, connectivity, latitude, longitude, etc.).
3. **Trips:** Trips are collections of legs that take place on a given day.
    Note that this day does not necessarily start (and end) at minight,
    but can start (and end) at an hour that is more logical/significant for the
    user (it could for example be 06:00 for car drivers). This day_start_hour
    parameter is universal for all trips

This module also includes two functions to declare a chosen class, and
to run that function for all class types.
'''

import datetime
import math
import typing as ty

import numpy as np
import pandas as pd
from box import Box
from ETS_CookBook import ETS_CookBook as cook

try:
    import run_time  # type: ignore

    # We need to ignore the type because mypy has its own search path for
    # imports and does not resolve imports exactly as Python does and it
    # isn't able to find the module.
    # https://stackoverflow.com/questions/68695851/mypy-cannot-find-implementation-or-library-stub-for-module
except ModuleNotFoundError:
    from ChaProEV import run_time  # type: ignore
# So that it works both as a standalone (1st) and as a package (2nd)
# We need to add to type: ignore thing to avoid MypY thinking
# we are importing again


def get_trip_charging_sessions(
    end_locations_of_legs: ty.List[str],
    start_probabilities: ty.List[float],
    time_between_legs_used: ty.List[float],
    leg_driving_times: ty.List[float],
    leg_distances: ty.List[float],
    weighted_leg_distances: ty.List[float],
    scenario: Box,
    general_parameters: Box,
) -> ty.List:
    trip_charging_sessions: ty.List[TripChargingSession] = []
    charging_session_resolution: int = scenario.charging_sessions.resolution
    vehicle_electricity_consumption: float = (
        scenario.vehicle.base_consumption_per_km.electricity_kWh
    )

    for travelling_group_start_slot, travelling_group_size in enumerate(
        start_probabilities
    ):
        if travelling_group_size > 0:
            number_of_legs: int = len(leg_driving_times)
            if number_of_legs > 0:
                session_first_start: float = travelling_group_start_slot
                for leg_index, (
                    leg_destination,
                    time_after_leg,
                    time_driving,
                    distance,
                    weighted_distance,
                ) in enumerate(
                    zip(
                        end_locations_of_legs,
                        time_between_legs_used,
                        leg_driving_times,
                        leg_distances,
                        weighted_leg_distances,
                    )
                ):

                    session_first_start += time_driving
                    session_location: str = leg_destination
                    session_size: float = (
                        travelling_group_size / charging_session_resolution
                    )
                    incoming_consumption: float = (
                        distance * vehicle_electricity_consumption
                    )
                    if leg_index == number_of_legs - 1:
                        next_leg_consumption: float = 0
                        session_duration: float = (
                            general_parameters.time.HOURS_IN_A_DAY
                        ) - session_first_start
                    else:
                        next_leg_consumption = (
                            leg_distances[leg_index + 1]
                            * vehicle_electricity_consumption
                        )
                        session_duration = time_after_leg

                    location_connectivity: float = scenario.locations[
                        session_location
                    ].connectivity
                    location_charging_power_to_vehicle: float = (
                        scenario.locations[session_location].charging_power
                        * location_connectivity
                    )
                    location_charging_power_from_network: float = (
                        scenario.locations[session_location].charging_power
                        * location_connectivity
                        / scenario.locations[
                            session_location
                        ].charger_efficiency
                    )
                    location_discharge_power_from_vehicle: float = (
                        scenario.locations[
                            session_location
                        ].vehicle_discharge_power
                        * location_connectivity
                    )
                    location_discharge_power_to_network: float = (
                        scenario.locations[
                            session_location
                        ].vehicle_discharge_power
                        * location_connectivity
                        * scenario.locations[
                            session_location
                        ].proportion_of_discharge_to_network
                    )

                    for sessions_substep_index in range(
                        charging_session_resolution
                    ):
                        substep_shift: float = 1 - 1 / (
                            sessions_substep_index + 1
                        )
                        session_start: float = (
                            session_first_start + substep_shift
                        )
                        session_end: float = session_start + session_duration
                        substep_charging_session: TripChargingSession = (
                            TripChargingSession(
                                session_location,
                                session_size,
                                session_start,
                                session_end,
                                incoming_consumption,
                                next_leg_consumption,
                                location_connectivity,
                                location_charging_power_to_vehicle,
                                location_charging_power_from_network,
                                location_discharge_power_from_vehicle,
                                location_discharge_power_to_network,
                            )
                        )
                        trip_charging_sessions.append(substep_charging_session)

    return trip_charging_sessions


def mobility_matrix_to_run_mobility_matrix(
    matrix_to_expand,  # It is a DataFRame, but MyPy seems to have issues
    # with DataFrmes with a MultiIndex
    leg_tuples: ty.List[ty.Tuple[str, str]],
    run_index: pd.MultiIndex,
    run_time_tags: pd.DatetimeIndex,
    start_hour: int,
    scenario: Box,
    general_parameters: Box,
) -> pd.DataFrame:
    run_matrix: pd.DataFrame = pd.DataFrame(
        columns=matrix_to_expand.columns, index=run_index
    )
    run_matrix = run_matrix.sort_index()
    for leg_tuple in leg_tuples:
        cloned_matrix: pd.DataFrame = run_time.from_day_to_run(
            matrix_to_expand.loc[
                (leg_tuple[0], leg_tuple[1]), matrix_to_expand.columns
            ],
            run_time_tags,
            start_hour,
            scenario,
            general_parameters,
        )
        for matrix_quantity in matrix_to_expand.columns:
            run_matrix.at[(leg_tuple[0], leg_tuple[1]), matrix_quantity] = (
                cloned_matrix[matrix_quantity].values
            )
    return run_matrix


def get_slot_split(
    percent_in_first_slot, travelling_group_size
) -> ty.Tuple[float, float, float]:

    percent_in_next_slot: float = 1 - percent_in_first_slot

    first_slot_impact_in_first_slot: float = (
        percent_in_first_slot * percent_in_first_slot / 2
    )

    first_slot_impact_in_second_slot: float = (
        percent_in_first_slot - first_slot_impact_in_first_slot
    )
    second_slot_impact_in_second_slot: float = (
        percent_in_next_slot * (1 + percent_in_first_slot) / 2
    )
    second_slot_impact_in_third_slot: float = (
        percent_in_next_slot - second_slot_impact_in_second_slot
    )

    first_slot_impact: float = (
        first_slot_impact_in_first_slot * travelling_group_size
    )

    second_slot_impact: float = (
        first_slot_impact_in_second_slot + second_slot_impact_in_second_slot
    ) * travelling_group_size
    third_slot_impact: float = (
        second_slot_impact_in_third_slot * travelling_group_size
    )

    return first_slot_impact, second_slot_impact, third_slot_impact


def compute_travel_impact(
    impacted_type: str,
    mobility_matrix,  # It is a DataFrame, but MyPy has issues with MultiIndex
    # DataFrames (and using loc on them)
    travelling_group_size: float,
    travelling_group_first_slot: int,
    percent_in_first_slot: float,
    leg_origin: str,
    leg_destination: str,
    distance: float,
    weighted_distance: float,
    battery_space_shifts: ty.Dict,
    vehicle_electricity_consumption: float,
) -> pd.DataFrame:

    leg_consumption: float = vehicle_electricity_consumption * distance
    weighted_leg_consumption: float = (
        vehicle_electricity_consumption * weighted_distance
    )

    first_slot_impact, second_slot_impact, third_slot_impact = get_slot_split(
        percent_in_first_slot, travelling_group_size
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        f'{impacted_type} impact',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][f'{impacted_type} impact']
        + first_slot_impact
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        f'{impacted_type} amount',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][f'{impacted_type} amount']
        + travelling_group_size * percent_in_first_slot
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        f'{impacted_type} impact kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][f'{impacted_type} impact kilometers']
        + first_slot_impact * distance
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        f'{impacted_type} kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][f'{impacted_type} kilometers']
        + travelling_group_size * percent_in_first_slot * distance
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        f'{impacted_type} impact weighted kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][f'{impacted_type} impact weighted kilometers']
        + first_slot_impact * weighted_distance
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        f'{impacted_type} weighted kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][f'{impacted_type} weighted kilometers']
        + travelling_group_size * percent_in_first_slot * weighted_distance
    )

    battery_space_shifts[(impacted_type, 'Impact')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Impact')].loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][leg_consumption]
        + first_slot_impact
    )
    battery_space_shifts[(impacted_type, 'Amount')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Amount')].loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][leg_consumption]
        + travelling_group_size * percent_in_first_slot
    )
    battery_space_shifts[(impacted_type, 'Impact Weighted')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        weighted_leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Impact Weighted')].loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][weighted_leg_consumption]
        + first_slot_impact
    )
    battery_space_shifts[(impacted_type, 'Amount Weighted')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot),
        weighted_leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Amount Weighted')].loc[
            leg_origin, leg_destination, travelling_group_first_slot
        ][weighted_leg_consumption]
        + travelling_group_size * percent_in_first_slot
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        f'{impacted_type} impact',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][f'{impacted_type} impact']
        + second_slot_impact
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        f'{impacted_type} amount',
    ] = mobility_matrix.loc[
        leg_origin, leg_destination, travelling_group_first_slot + 1
    ][
        f'{impacted_type} amount'
    ] + travelling_group_size * (
        1 - percent_in_first_slot
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        f'{impacted_type} impact kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][f'{impacted_type} impact kilometers']
        + second_slot_impact * distance
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        f'{impacted_type} kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][f'{impacted_type} kilometers']
        + travelling_group_size * (1 - percent_in_first_slot) * distance
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        f'{impacted_type} impact weighted kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][f'{impacted_type} impact weighted kilometers']
        + second_slot_impact * weighted_distance
    )
    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        f'{impacted_type} weighted kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][f'{impacted_type} weighted kilometers']
        + travelling_group_size
        * (1 - percent_in_first_slot)
        * weighted_distance
    )

    battery_space_shifts[(impacted_type, 'Impact')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Impact')].loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][leg_consumption]
        + second_slot_impact
    )
    battery_space_shifts[(impacted_type, 'Amount')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        leg_consumption,
    ] = battery_space_shifts[(impacted_type, 'Amount')].loc[
        leg_origin, leg_destination, travelling_group_first_slot + 1
    ][
        leg_consumption
    ] + travelling_group_size * (
        1 - percent_in_first_slot
    )
    battery_space_shifts[(impacted_type, 'Impact Weighted')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        weighted_leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Impact Weighted')].loc[
            leg_origin, leg_destination, travelling_group_first_slot + 1
        ][weighted_leg_consumption]
        + second_slot_impact
    )
    battery_space_shifts[(impacted_type, 'Amount Weighted')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 1),
        weighted_leg_consumption,
    ] = battery_space_shifts[(impacted_type, 'Amount Weighted')].loc[
        leg_origin, leg_destination, travelling_group_first_slot + 1
    ][
        weighted_leg_consumption
    ] + travelling_group_size * (
        1 - percent_in_first_slot
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 2),
        f'{impacted_type} impact',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 2
        ][f'{impacted_type} impact']
        + third_slot_impact
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 2),
        f'{impacted_type} impact kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 2
        ][f'{impacted_type} impact kilometers']
        + third_slot_impact * distance
    )

    mobility_matrix.loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 2),
        f'{impacted_type} impact weighted kilometers',
    ] = (
        mobility_matrix.loc[
            leg_origin, leg_destination, travelling_group_first_slot + 2
        ][f'{impacted_type} impact weighted kilometers']
        + third_slot_impact * weighted_distance
    )

    battery_space_shifts[(impacted_type, 'Impact')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 2),
        leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Impact')].loc[
            leg_origin, leg_destination, travelling_group_first_slot + 2
        ][leg_consumption]
        + third_slot_impact
    )

    battery_space_shifts[(impacted_type, 'Impact Weighted')].loc[
        (leg_origin, leg_destination, travelling_group_first_slot + 2),
        weighted_leg_consumption,
    ] = (
        battery_space_shifts[(impacted_type, 'Impact Weighted')].loc[
            leg_origin, leg_destination, travelling_group_first_slot + 2
        ][weighted_leg_consumption]
        + third_slot_impact
    )

    return mobility_matrix


def get_travelling_group_travel_impact(
    mobility_matrix: pd.DataFrame,
    travelling_group_size: float,
    travelling_group_start_slot: int,
    time_between_legs_used: ty.List[float],
    leg_driving_times: ty.List[float],
    start_locations_of_legs: ty.List[str],
    end_locations_of_legs: ty.List[str],
    leg_distances: ty.List[float],
    leg_weighted_distances: ty.List[float],
    battery_space_shifts: ty.Dict[ty.Tuple[str, str], pd.DataFrame],
    vehicle_electricity_consumption: float,
    HOURS_IN_A_DAY: int,
) -> pd.DataFrame:

    travelling_group_first_slot: int = travelling_group_start_slot
    percent_in_first_slot: float = 1
    if travelling_group_first_slot + 2 >= HOURS_IN_A_DAY:

        print('Trip extending beyond day end. Check your entry')
        exit()
    # The trips start uniformly within the first slot

    for leg_index, (
        leg_origin,
        leg_destination,
        time_after_leg,
        time_driving,
        distance,
        weighted_distance,
    ) in enumerate(
        zip(
            start_locations_of_legs,
            end_locations_of_legs,
            time_between_legs_used,
            leg_driving_times,
            leg_distances,
            leg_weighted_distances,
        )
    ):
        # We begin with the impact of departures

        mobility_matrix = compute_travel_impact(
            'Departures',
            mobility_matrix,
            travelling_group_size,
            travelling_group_first_slot,
            percent_in_first_slot,
            leg_origin,
            leg_destination,
            distance,
            weighted_distance,
            battery_space_shifts,
            vehicle_electricity_consumption,
        )

        # We update the slots of the group
        travelling_group_first_slot += math.floor(time_driving)
        percent_in_first_slot -= time_driving % 1

        if percent_in_first_slot <= 0:
            travelling_group_first_slot += 1
            percent_in_first_slot = 1 + percent_in_first_slot

        if travelling_group_first_slot + 2 >= HOURS_IN_A_DAY:

            print('Trip extending beyond day end. Check your entry')
            exit()

        # We now look at the impact of arrivals
        mobility_matrix = compute_travel_impact(
            'Arrivals',
            mobility_matrix,
            travelling_group_size,
            travelling_group_first_slot,
            percent_in_first_slot,
            leg_origin,
            leg_destination,
            distance,
            weighted_distance,
            battery_space_shifts,
            vehicle_electricity_consumption,
        )

        # We update the slots of the group
        travelling_group_first_slot += math.floor(time_after_leg)
        percent_in_first_slot -= time_after_leg % 1

        if percent_in_first_slot <= 0:
            travelling_group_first_slot += 1
            percent_in_first_slot = 1 + percent_in_first_slot

        if travelling_group_first_slot + 2 >= HOURS_IN_A_DAY:

            print('Trip extending beyond day end. Check your entry')
            exit()

    return mobility_matrix


def get_location_split_and_impact_of_departures_and_arrivals(
    location_names: ty.List[str],
    location_split: pd.DataFrame,
    mobility_matrix,  # It is a DataFrame, but Mypy has issues with
    # assigning MultiIndex DataFrames
    start_probabilities: ty.List[float],
    time_between_legs: ty.List[float],
    leg_driving_times: ty.List[float],
    start_locations_of_legs: ty.List[str],
    end_locations_of_legs: ty.List[str],
    leg_distances: ty.List[float],
    weighted_leg_distances: ty.List[float],
    battery_space_shifts: ty.Dict[ty.Tuple[str, str], pd.DataFrame],
    vehicle_electricity_consumption: float,
    HOURS_IN_A_DAY: int,
) -> ty.Tuple[pd.DataFrame, pd.DataFrame]:
    if len(leg_driving_times) > 0:
        # If there are no legs, the arrays stay full of zeroes
        dummy_time_between_legs: float = 0
        # This dummy value is added so that we can iterate through a
        # zip of driving times and times between legs that have the
        # same length as the amount of legs
        time_between_legs_used: ty.List[float] = time_between_legs.copy()
        time_between_legs_used.append(dummy_time_between_legs)
        for travelling_group_start_slot, travelling_group_size in enumerate(
            start_probabilities
        ):
            if travelling_group_size > 0:
                mobility_matrix = get_travelling_group_travel_impact(
                    mobility_matrix,
                    travelling_group_size,
                    travelling_group_start_slot,
                    time_between_legs_used,
                    leg_driving_times,
                    start_locations_of_legs,
                    end_locations_of_legs,
                    leg_distances,
                    weighted_leg_distances,
                    battery_space_shifts,
                    vehicle_electricity_consumption,
                    HOURS_IN_A_DAY,
                )

        possible_origins: ty.List[str] = list(
            set(mobility_matrix.index.get_level_values('From'))
        )

        for start_location in possible_origins:
            possible_destinations: ty.List[str] = list(
                set(
                    (
                        mobility_matrix.loc[
                            start_location
                        ].index.get_level_values('To')
                    )
                )
            )
            for destination in possible_destinations:
                location_split[destination] = (
                    location_split[destination].values
                    + mobility_matrix.loc[start_location, destination][
                        'Arrivals impact'
                    ]
                    .cumsum()
                    .values
                )
                location_split[start_location] = (
                    location_split[start_location].values
                    - mobility_matrix.loc[start_location, destination][
                        'Departures impact'
                    ]
                    .cumsum()
                    .values
                )

    return location_split, mobility_matrix


class Leg:
    '''
    This class defines the legs and their properties, from a scenario
    file that contains a list of instances and their properties.
    Legs are point-to-point vehicle movements (i.e. movements where
    the vehicle goes from a start location and ends/stops at an end location).
    '''

    class_name: str = 'legs'

    def __init__(
        leg, name: str, scenario: Box, general_parameters: Box
    ) -> None:
        leg.name: str = name

        leg_parameters: Box = scenario.legs[name]
        leg.vehicle: str = leg_parameters.vehicle
        leg.distance: float = leg_parameters.distance
        leg.duration: float = leg_parameters.duration
        leg.hour_in_day_factors: ty.List[float] = (
            leg_parameters.hour_in_day_factors
        )
        locations: Box = leg_parameters.locations
        leg.start_location: str = locations.start
        leg.end_location: str = locations.end
        road_type_parameters: Box = leg_parameters.road_type_mix
        leg.road_type_mix: ty.List[float] = road_type_parameters.mix


class Location:
    '''
    This class defines the locations where the vehicles are
    and their properties,  from a scenario
    file that contains a list of instances and their properties.
    '''

    class_name: str = 'locations'

    def __init__(
        location, name: str, scenario: Box, general_parameters: Box
    ) -> None:
        location.name: str = name

        this_location_parameters: Box = scenario.locations[name]
        location.vehicle: str = this_location_parameters.vehicle
        location.connectivity: float = this_location_parameters.connectivity
        location.charging_power: float = (
            this_location_parameters.charging_power
        )
        location.latitude: float = this_location_parameters.latitude
        location.longitude: float = this_location_parameters.longitude
        location.base_charging_price: float = (
            this_location_parameters.base_charging_price
        )
        location.charging_desirability: float = (
            this_location_parameters.charging_desirability
        )


class Trip:
    '''
    This class defines the  trips and their properties, from a scenario
    file that contains a list of instances and their properties.
    Trips are collections of legs that take place on a given day.
    Note that this day does not necessarily start (and end) at minight,
    but can start (and end) at an hour that is more logical/significant for
    the
    vehicle user (it could for example be 06:00 for car drivers).
    This day_start_hour
    parameter is universal for all trips
    This value is set in the scenario files.
    '''

    class_name: str = 'trips'

    def __init__(
        trip, name: str, scenario: Box, general_parameters: Box
    ) -> None:
        trip.name: str = name

        trip_parameters: Box = scenario.trips[name]
        trip.legs: ty.List[str] = trip_parameters.legs
        trip.time_between_legs: ty.List[float] = (
            trip_parameters.time_between_legs
        )
        trip.percentage_station_users: float = (
            trip_parameters.percentage_station_users
        )

        trip.start_probabilities: ty.List[float] = (
            trip_parameters.start_probabilities
        )
        trip.repeated_sequence: ty.List[str] = (
            trip_parameters.repeated_sequence
        )
        trip.repetition_amounts: ty.List[int] = (
            trip_parameters.repetition_amounts
        )
        trip.time_between_repetitions: ty.List[float] = (
            trip_parameters.time_between_repetitions
        )

        trip.day_start_hour: int = scenario.mobility_module.day_start_hour
        trip.leg_driving_times: ty.List[float] = [
            scenario.legs[leg_name].duration for leg_name in trip.legs
        ]

        trip_legs_store: ty.List[str] = trip.legs.copy()

        time_between_legs_store: ty.List[float] = trip.time_between_legs.copy()
        leg_driving_times_store: ty.List[float] = trip.leg_driving_times.copy()
        if len(trip.repeated_sequence) > 0:
            trip.legs = []
            trip.time_between_legs = []
            trip.leg_driving_times = []
            repetition_iteration_index: int = 0
            for leg_index, leg in enumerate(trip_legs_store):

                if leg not in trip.repeated_sequence:
                    trip.legs.append(leg)
                    if leg_index < len(trip_legs_store) - 1:
                        trip.time_between_legs.append(
                            time_between_legs_store[leg_index]
                        )
                    trip.leg_driving_times.append(
                        leg_driving_times_store[leg_index]
                    )
                elif leg == trip.repeated_sequence[0]:
                    for repetition in range(
                        trip.repetition_amounts[repetition_iteration_index] + 1
                    ):
                        for leg_to_add_index, leg_to_add in enumerate(
                            trip.repeated_sequence
                        ):
                            trip.legs.append(leg_to_add)
                            if (
                                leg_to_add_index
                                < len(trip.repeated_sequence) - 1
                            ):
                                trip.time_between_legs.append(
                                    time_between_legs_store[leg_index]
                                )
                            trip.leg_driving_times.append(
                                leg_driving_times_store[leg_index]
                            )
                        if (
                            repetition
                            < trip.repetition_amounts[
                                repetition_iteration_index
                            ]
                        ):
                            trip.time_between_legs.append(
                                trip.time_between_repetitions[
                                    repetition_iteration_index
                                ]
                            )

                        else:

                            add_final_time_between_legs: bool = True

                            if leg_index == (
                                len(trip_legs_store)
                                - len(trip.repeated_sequence)
                            ):

                                if (
                                    trip_legs_store[-1]
                                    == trip.repeated_sequence[-1]
                                ):
                                    add_final_time_between_legs = False
                                    # If the last leg of the trip
                                    # is also the last leg of the repeated
                                    # sequence, we don't need to add
                                    # the time after it (since it is the
                                    # last leg of the trip).

                            if add_final_time_between_legs:
                                trip.time_between_legs.append(
                                    time_between_legs_store[
                                        leg_index
                                        + len(trip.repeated_sequence)
                                        - 1
                                    ]
                                )

                    repetition_iteration_index += 1

        trip.start_locations_of_legs: ty.List[str] = [
            scenario.legs[leg_name].locations.start for leg_name in trip.legs
        ]
        trip.end_locations_of_legs: ty.List[str] = [
            scenario.legs[leg_name].locations.end for leg_name in trip.legs
        ]
        trip.location_names: ty.List[str] = sorted(
            list(
                set(trip.start_locations_of_legs + trip.end_locations_of_legs)
            )
        )
        if trip.name.startswith('stay_put_'):
            stay_put_location: str = trip.name.split('stay_put_')[1]
            trip.location_names = [stay_put_location]
        # sorted so that the order is always the same

        trip.leg_distances: ty.List[float] = [
            scenario.legs[leg_name].distance for leg_name in trip.legs
        ]

        trip.weighted_leg_distances: ty.List[float] = []
        for leg_name in trip.legs:
            road_type_mix: np.ndarray = np.array(
                scenario.legs[leg_name].road_type_mix.mix
            )
            road_type_weights: np.ndarray = np.array(
                scenario.transport_factors.weights
            )
            road_type_factor: float = float(
                sum(road_type_mix * road_type_weights)
            )
            leg_distance: float = scenario.legs[leg_name].distance
            leg_weighted_distance: float = leg_distance * road_type_factor
            trip.weighted_leg_distances.append(leg_weighted_distance)

        trip.unique_leg_distances: ty.List[float] = list(
            set(trip.leg_distances)
        )
        trip.unique_weighted_leg_distances: ty.List[float] = list(
            set(trip.weighted_leg_distances)
        )
        vehicle_electricity_consumption: float = (
            scenario.vehicle.base_consumption_per_km.electricity_kWh
        )

        trip.unique_leg_consumptions: ty.List[float] = [
            unique_distance * vehicle_electricity_consumption
            for unique_distance in trip.unique_leg_distances
        ]

        trip.unique_weighted_leg_consumptions: ty.List[float] = [
            unique_distance * vehicle_electricity_consumption
            for unique_distance in trip.unique_weighted_leg_distances
        ]

        # We want to create a mobility matrix for the trip. This matrix will
        # have start and end locations (plus hour in day, starting
        # at day start) as an index, and departures, arrivals as columns (
        # each with amounts, distances, weighted distances)
        HOURS_IN_A_DAY: int = general_parameters.time.HOURS_IN_A_DAY
        parameters_of_legs: Box = scenario.legs
        unique_legs: ty.List[str] = []
        for trip_leg in trip.legs:
            if trip_leg not in unique_legs:
                unique_legs.append(trip_leg)

        leg_tuples: ty.List[ty.Tuple[str, str]] = [
            (
                parameters_of_legs[trip_leg].locations.start,
                parameters_of_legs[trip_leg].locations.end,
            )
            for trip_leg in unique_legs
        ]

        # We only want the start and end locations that are in the legs
        mobility_index_tuples: ty.List[ty.Tuple[str, str, int]] = [
            (leg_tuple[0], leg_tuple[1], hour_number)
            for leg_tuple in leg_tuples
            for hour_number in range(HOURS_IN_A_DAY)
        ]

        mobility_index: pd.MultiIndex = pd.MultiIndex.from_tuples(
            mobility_index_tuples,
            names=['From', 'To', 'Hour number (from day start)'],
        )
        mobility_quantities: ty.List[str] = (
            scenario.mobility_module.mobility_quantities
        )

        trip.mobility_matrix: pd.DataFrame = pd.DataFrame(
            np.zeros((len(mobility_index), len(mobility_quantities))),
            columns=mobility_quantities,
            index=mobility_index,
        )

        trip.mobility_matrix = trip.mobility_matrix.sort_index()

        # We also want to store the battery space shifts true to legs.
        # We track this as departure and arrivals (including a version with
        # impact versions).
        base_dataframe_for_battery_space_shifts: pd.DataFrame = pd.DataFrame(
            np.zeros((len(mobility_index), len(trip.unique_leg_consumptions))),
            columns=trip.unique_leg_consumptions,
            index=mobility_index,
        )
        base_dataframe_for_weighted_battery_space_shifts: pd.DataFrame = (
            pd.DataFrame(
                np.zeros(
                    (
                        len(mobility_index),
                        len(trip.unique_weighted_leg_consumptions),
                    )
                ),
                columns=trip.unique_weighted_leg_consumptions,
                index=mobility_index,
            )
        )
        base_dataframe_for_weighted_battery_space_shifts = (
            base_dataframe_for_weighted_battery_space_shifts.sort_index()
        )
        trip.battery_space_shifts_departures: pd.DataFrame = (
            base_dataframe_for_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_departures = (
            trip.battery_space_shifts_departures.sort_index()
        )
        trip.battery_space_shifts_departures_impact: pd.DataFrame = (
            base_dataframe_for_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_departures_impact = (
            trip.battery_space_shifts_departures_impact.sort_index()
        )
        trip.battery_space_shifts_arrivals: pd.DataFrame = (
            base_dataframe_for_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_arrivals = (
            trip.battery_space_shifts_arrivals.sort_index()
        )
        trip.battery_space_shifts_arrivals_impact: pd.DataFrame = (
            base_dataframe_for_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_arrivals_impact = (
            trip.battery_space_shifts_arrivals_impact.sort_index()
        )
        trip.battery_space_shifts_departures_weighted: pd.DataFrame = (
            base_dataframe_for_weighted_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_departures_weighted = (
            trip.battery_space_shifts_departures_weighted.sort_index()
        )
        trip.battery_space_shifts_departures_impact_weighted: pd.DataFrame = (
            base_dataframe_for_weighted_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_departures_impact_weighted = (
            trip.battery_space_shifts_departures_impact_weighted.sort_index()
        )
        trip.battery_space_shifts_arrivals_weighted: pd.DataFrame = (
            base_dataframe_for_weighted_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_arrivals_weighted = (
            trip.battery_space_shifts_arrivals_weighted.sort_index()
        )
        trip.battery_space_shifts_arrivals_impact_weighted: pd.DataFrame = (
            base_dataframe_for_weighted_battery_space_shifts.copy()
        )
        trip.battery_space_shifts_arrivals_impact_weighted = (
            trip.battery_space_shifts_arrivals_impact_weighted.sort_index()
        )
        trip.battery_space_shifts: ty.Dict[
            ty.Tuple[str, str], pd.DataFrame
        ] = {}
        trip.battery_space_shifts[('Departures', 'Amount')] = (
            trip.battery_space_shifts_departures
        )
        trip.battery_space_shifts[('Departures', 'Impact')] = (
            trip.battery_space_shifts_departures_impact
        )
        trip.battery_space_shifts[('Departures', 'Amount Weighted')] = (
            trip.battery_space_shifts_departures_weighted
        )
        trip.battery_space_shifts[('Departures', 'Impact Weighted')] = (
            trip.battery_space_shifts_departures_impact_weighted
        )
        trip.battery_space_shifts[('Arrivals', 'Amount')] = (
            trip.battery_space_shifts_arrivals
        )
        trip.battery_space_shifts[('Arrivals', 'Impact')] = (
            trip.battery_space_shifts_arrivals_impact
        )
        trip.battery_space_shifts[('Arrivals', 'Amount Weighted')] = (
            trip.battery_space_shifts_arrivals_weighted
        )
        trip.battery_space_shifts[('Arrivals', 'Impact Weighted')] = (
            trip.battery_space_shifts_arrivals_impact_weighted
        )

        # We want to track where the vehicle is
        trip.location_split: pd.DataFrame = pd.DataFrame(
            np.zeros((HOURS_IN_A_DAY, len(trip.location_names))),
            columns=trip.location_names,
            index=range(HOURS_IN_A_DAY),
        )
        trip.location_split.index.name = 'Hour in day (from day start)'

        if trip.name.startswith('stay_put_'):
            stay_put_location = trip.name.split('stay_put_')[1]
            trip.location_split[stay_put_location] = 1
        else:
            first_leg: str = trip.legs[0]
            initial_location: str = scenario.legs[first_leg].locations.start
            trip.location_split[initial_location] = 1

        trip.location_split, trip.mobility_matrix = (
            get_location_split_and_impact_of_departures_and_arrivals(
                trip.location_names,
                trip.location_split,
                trip.mobility_matrix,
                trip.start_probabilities,
                trip.time_between_legs,
                trip.leg_driving_times,
                trip.start_locations_of_legs,
                trip.end_locations_of_legs,
                trip.leg_distances,
                trip.weighted_leg_distances,
                trip.battery_space_shifts,
                vehicle_electricity_consumption,
                HOURS_IN_A_DAY,
            )
        )

        # We can also get the percentage driving

        trip.percentage_driving: pd.Series[float] = (
            1 - trip.location_split.sum(axis=1)
        )

        # We also get the connectivity:
        trip.connectivity_per_location: pd.DataFrame = (
            trip.location_split.copy()
        )
        trip.maximal_delivered_power_per_location: pd.DataFrame = (
            trip.location_split.copy()
        )  # How much the network can discharge
        trip.maximal_received_power_per_location: pd.DataFrame = (
            trip.location_split.copy()
        )  # How much of tht the vehicles can receive
        trip.vehicle_discharge_power_per_location: pd.DataFrame = (
            trip.location_split.copy()
        )
        trip.discharge_power_to_network_per_location: pd.DataFrame = (
            trip.location_split.copy()
        )

        for location_name in trip.location_names:
            this_location_parameters: Box = scenario.locations[location_name]
            location_connectivity: float = (
                this_location_parameters.connectivity
            )
            charging_power: float = this_location_parameters.charging_power
            charger_efficiency: float = (
                this_location_parameters.charger_efficiency
            )
            this_location_maximal_delivered_power: float = (
                location_connectivity * charging_power / charger_efficiency
            )
            this_location_maximal_received_power: float = (
                location_connectivity * charging_power
            )
            this_location_vehicle_discharge_power: float = (
                this_location_parameters.vehicle_discharge_power
            )
            this_location_proportion_of_discharge_to_network: float = (
                this_location_parameters.proportion_of_discharge_to_network
            )
            this_location_discharge_power_to_network: float = (
                this_location_vehicle_discharge_power
                * this_location_proportion_of_discharge_to_network
            )

            trip.connectivity_per_location[location_name] = (
                trip.connectivity_per_location[location_name]
                * location_connectivity
            )
            trip.maximal_delivered_power_per_location[location_name] = (
                trip.maximal_delivered_power_per_location[location_name]
                * this_location_maximal_delivered_power
            )
            trip.maximal_received_power_per_location[location_name] = (
                trip.maximal_received_power_per_location[location_name]
                * this_location_maximal_received_power
            )

            trip.vehicle_discharge_power_per_location[location_name] = (
                trip.vehicle_discharge_power_per_location[location_name]
                * this_location_vehicle_discharge_power
            )
            trip.discharge_power_to_network_per_location[location_name] = (
                trip.discharge_power_to_network_per_location[location_name]
                * this_location_discharge_power_to_network
            )
        trip.connectivity: pd.Series = trip.connectivity_per_location.sum(
            axis=1
        )
        trip.maximal_delivered_power: pd.Series = (
            trip.maximal_delivered_power_per_location.sum(axis=1)
        )
        trip.maximal_received_power: pd.Series = (
            trip.maximal_received_power_per_location.sum(axis=1)
        )
        trip.vehicle_discharge_power: pd.Series = (
            trip.vehicle_discharge_power_per_location.sum(axis=1)
        )
        trip.discharge_power_to_network: pd.Series = (
            trip.discharge_power_to_network_per_location.sum(axis=1)
        )
        dummy_time_between_legs: float = 0
        # This dummy value is added so that we can iterate through a
        # zip of driving times and times between legs that have the
        # same length as the amount of legs
        time_between_legs_used: ty.List[float] = trip.time_between_legs.copy()
        time_between_legs_used.append(dummy_time_between_legs)

        trip.charging_sessions: ty.List[TripChargingSession] = (
            get_trip_charging_sessions(
                trip.end_locations_of_legs,
                trip.start_probabilities,
                time_between_legs_used,
                trip.leg_driving_times,
                trip.leg_distances,
                trip.weighted_leg_distances,
                scenario,
                general_parameters,
            )
        )

        # We now can create a mobility matrix for the whole run
        run_time_tags: pd.DatetimeIndex = run_time.get_time_range(
            scenario, general_parameters
        )[0]

        # We only want the start and end locations that are in the legs
        run_mobility_index_tuples: ty.List[
            ty.Tuple[str, str, datetime.datetime]
        ] = [
            (leg_tuple[0], leg_tuple[1], time_tag)
            for leg_tuple in leg_tuples
            for time_tag in run_time_tags
        ]

        mobility_index_names: ty.List[str] = (
            scenario.mobility_module.mobility_index_names
        )
        run_mobility_index: pd.MultiIndex = pd.MultiIndex.from_tuples(
            run_mobility_index_tuples, names=mobility_index_names
        )

        trip.run_mobility_matrix: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.mobility_matrix,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        ).astype(float)

        trip.run_battery_space_shifts_departures: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.battery_space_shifts_departures,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_battery_space_shifts_departures_impact: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.battery_space_shifts_departures_impact,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_battery_space_shifts_arrivals: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.battery_space_shifts_arrivals,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_battery_space_shifts_arrivals_impact: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.battery_space_shifts_arrivals_impact,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )

        trip.run_battery_space_shifts_departures_weighted: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.battery_space_shifts_departures_weighted,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_battery_space_shifts_departures_impact_weighted: (
            pd.DataFrame
        ) = mobility_matrix_to_run_mobility_matrix(
            trip.battery_space_shifts_departures_impact_weighted,
            leg_tuples,
            run_mobility_index,
            run_time_tags,
            trip.day_start_hour,
            scenario,
            general_parameters,
        )
        trip.run_battery_space_shifts_arrivals_weighted: pd.DataFrame = (
            mobility_matrix_to_run_mobility_matrix(
                trip.battery_space_shifts_arrivals_weighted,
                leg_tuples,
                run_mobility_index,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_battery_space_shifts_arrivals_impact_weighted: (
            pd.DataFrame
        ) = mobility_matrix_to_run_mobility_matrix(
            trip.battery_space_shifts_arrivals_impact_weighted,
            leg_tuples,
            run_mobility_index,
            run_time_tags,
            trip.day_start_hour,
            scenario,
            general_parameters,
        )

        # We also do this for some other quantities
        day_start_hour: int = scenario.mobility_module.day_start_hour
        trip.run_location_split: pd.DataFrame = run_time.from_day_to_run(
            trip.location_split,
            run_time_tags,
            day_start_hour,
            scenario,
            general_parameters,
        )

        trip.run_connectivity_per_location: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.connectivity_per_location,
                run_time_tags,
                day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_maximal_delivered_power_per_location: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.maximal_delivered_power_per_location,
                run_time_tags,
                day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_maximal_received_power_per_location: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.maximal_received_power_per_location,
                run_time_tags,
                day_start_hour,
                scenario,
                general_parameters,
            )
        )

        trip.run_vehicle_discharge_power_per_location: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.vehicle_discharge_power_per_location,
                run_time_tags,
                day_start_hour,
                scenario,
                general_parameters,
            )
        )

        trip.run_discharge_power_to_network_per_location: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.discharge_power_to_network_per_location,
                run_time_tags,
                day_start_hour,
                scenario,
                general_parameters,
            )
        )

        trip.run_percentage_driving: pd.Series = (
            1 - trip.run_location_split.sum(axis=1)
        )

        trip.run_connectivity: pd.Series = (
            trip.run_connectivity_per_location.sum(axis=1)
        )
        trip.run_maximal_delivered_power: pd.Series = (
            trip.run_maximal_delivered_power_per_location.sum(axis=1)
        )
        trip.run_maximal_received_power: pd.Series = (
            trip.run_maximal_received_power_per_location.sum(axis=1)
        )

        trip.run_vehicle_discharge_power: pd.Series = (
            trip.run_vehicle_discharge_power_per_location.sum(axis=1)
        )

        trip.run_discharge_power_to_network: pd.Series = (
            trip.run_discharge_power_to_network_per_location.sum(axis=1)
        )

        trip.next_leg_kilometers: pd.DataFrame = pd.DataFrame(
            np.zeros((HOURS_IN_A_DAY, len(trip.location_names))),
            columns=trip.location_names,
            index=range(HOURS_IN_A_DAY),
        )
        trip.next_leg_kilometers.index.name = 'Hour number (from day start)'
        # The standard version sets the kilometers in the time slot
        # before departure, this second, cumulative, version does so from
        # the moment the vehicles arrive
        # (or day start for the first leg of the trip)
        trip.next_leg_kilometers_cumulative: pd.DataFrame = (
            trip.next_leg_kilometers.copy()
        )

        for leg_index, leg_name in enumerate(trip.legs):

            leg_parameters = scenario.legs[leg_name]
            start_location = leg_parameters.locations.start
            end_location = leg_parameters.locations.end

            if leg_index == 0:
                previous_leg_arrivals_amount: ty.List[float] = []
                for hour_index in range(HOURS_IN_A_DAY - 1):
                    # We skip the last time slot, as this should wrap the trip
                    # and we are not looking into the next day
                    # with this apporach

                    # For the standard version, we look if the vehicle
                    # is set to depart in the next time
                    upcoming_departures_kilometers = trip.mobility_matrix.loc[
                        (start_location, end_location, hour_index + 1),
                        'Departures kilometers',
                    ]

                    trip.next_leg_kilometers.loc[
                        hour_index, start_location
                    ] = (
                        trip.next_leg_kilometers.loc[hour_index][
                            start_location
                        ]
                        + upcoming_departures_kilometers
                    )

                    # For the other version, we look at all future departures

                    all_future_departures_kilometers = (
                        trip.mobility_matrix.loc[
                            (
                                start_location,
                                end_location,
                                list(range(hour_index + 1, HOURS_IN_A_DAY)),
                            ),
                            'Departures kilometers',
                        ].sum()
                    )

                    trip.next_leg_kilometers_cumulative.loc[
                        hour_index, start_location
                    ] = (
                        trip.next_leg_kilometers_cumulative.loc[hour_index][
                            start_location
                        ]
                        + all_future_departures_kilometers
                    )

            else:
                for hour_index in range(HOURS_IN_A_DAY - 1):
                    # We skip the last time slot, as this should wrap the trip
                    # and we are not looking into the next day
                    # with this approach

                    # We do the same as for the first leg, but we need
                    # to limit that to the vehicles that already have
                    # arrived at the end location
                    upcoming_departures_kilometers_raw: float = (
                        trip.mobility_matrix['Departures kilometers'].loc[
                            start_location, end_location, hour_index + 1
                        ]
                    )

                    previous_arrivals_to_use: float = sum(
                        previous_leg_arrivals_amount[: hour_index + 1]
                    )
                    actual_upcoming_departures_kilometers = (
                        upcoming_departures_kilometers_raw
                        * previous_arrivals_to_use
                    )

                    trip.next_leg_kilometers.loc[
                        hour_index, start_location
                    ] = (
                        trip.next_leg_kilometers.loc[hour_index][
                            start_location
                        ]
                        + actual_upcoming_departures_kilometers
                    )

                    # Once again, the variant applies to all future departures

                    all_future_departures_kilometers_raw: float = (
                        trip.mobility_matrix['Departures kilometers']
                        .loc[start_location, end_location][
                            list(range(hour_index + 1, HOURS_IN_A_DAY))
                        ]
                        .sum()
                    )

                    actual_all_future_departures_kilometers = (
                        previous_arrivals_to_use
                        * all_future_departures_kilometers_raw
                    )

                    trip.next_leg_kilometers_cumulative.loc[
                        hour_index, start_location
                    ] = (
                        trip.next_leg_kilometers_cumulative.loc[hour_index][
                            start_location
                        ]
                        + actual_all_future_departures_kilometers
                    )

            previous_leg_arrivals_amount = trip.mobility_matrix[
                'Arrivals amount'
            ].loc[start_location, end_location]

        trip.next_leg_charge_to_vehicle: pd.DataFrame = (
            trip.next_leg_kilometers.copy()
        )
        trip.next_leg_charge_from_network: pd.DataFrame = (
            trip.next_leg_kilometers.copy()
        )
        trip.next_leg_charge_to_vehicle_cumulative: pd.DataFrame = (
            trip.next_leg_kilometers_cumulative.copy()
        )
        trip.next_leg_charge_from_network_cumulative: pd.DataFrame = (
            trip.next_leg_kilometers_cumulative.copy()
        )

        for location_name in trip.location_names:
            vehicle_consumption_per_km: float = (
                scenario.vehicle.base_consumption_per_km.electricity_kWh
            )
            charger_efficiency = scenario.locations[
                location_name
            ].charger_efficiency

            draw_from_network_per_km: float = (
                vehicle_consumption_per_km / charger_efficiency
            )
            trip.next_leg_charge_to_vehicle[location_name] = (
                trip.next_leg_charge_to_vehicle[location_name]
                * vehicle_consumption_per_km
            )
            trip.next_leg_charge_to_vehicle_cumulative[location_name] = (
                trip.next_leg_charge_to_vehicle_cumulative[location_name]
                * vehicle_consumption_per_km
            )
            trip.next_leg_charge_from_network[location_name] = (
                trip.next_leg_charge_from_network[location_name]
                * draw_from_network_per_km
            )
            trip.next_leg_charge_from_network_cumulative[location_name] = (
                trip.next_leg_charge_from_network_cumulative[location_name]
                * draw_from_network_per_km
            )

        trip.run_next_leg_kilometers: pd.DataFrame = run_time.from_day_to_run(
            trip.next_leg_kilometers,
            run_time_tags,
            trip.day_start_hour,
            scenario,
            general_parameters,
        )
        trip.run_next_leg_kilometers_cumulative: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.next_leg_kilometers_cumulative,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )

        trip.run_next_leg_charge_to_vehicle: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.next_leg_charge_to_vehicle,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_next_leg_charge_to_vehicle_cumulative: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.next_leg_charge_to_vehicle_cumulative,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )

        trip.run_next_leg_charge_from_network: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.next_leg_charge_from_network,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )
        trip.run_next_leg_charge_from_network_cumulative: pd.DataFrame = (
            run_time.from_day_to_run(
                trip.next_leg_charge_from_network_cumulative,
                run_time_tags,
                trip.day_start_hour,
                scenario,
                general_parameters,
            )
        )


class TripChargingSession:
    '''
    This class defines the charging sessions of an incoming group in a Trip.
    The quantities other than start time, end time, and location
    are proprtional to the size of the session (the group travelling).
    '''

    class_name: str = 'trip_charging_session'

    def __init__(
        charging_session,
        location_name: str,
        session_size: float,
        session_start: float,
        session_end: float,
        incoming_consumption: float,
        outgoing_consumption: float,
        location_connectivity: float,
        location_charging_power_to_vehicle: float,
        location_charging_power_from_network: float,
        location_discharge_power_from_vehicle: float,
        location_discharge_power_to_network: float,
    ) -> None:

        charging_session.start_time: float = session_start
        charging_session.end_time: float = session_end
        charging_session.location: str = location_name
        charging_session.previous_leg_consumption: float = (
            incoming_consumption * session_size
        )
        charging_session.next_leg_consumption: float = (
            outgoing_consumption * session_size
        )
        charging_session.connectivity: float = (
            location_connectivity * session_size
        )
        charging_session.power_to_vehicle: float = (
            location_charging_power_to_vehicle * session_size
        )
        charging_session.power_from_network: float = (
            location_charging_power_from_network * session_size
        )
        charging_session.power_from_vehicle: float = (
            location_discharge_power_from_vehicle * session_size
        )
        charging_session.power_to_network: float = (
            location_discharge_power_to_network * session_size
        )


def declare_class_instances(
    ChosenClass: ty.Type, scenario: Box, general_parameters: Box
) -> ty.List[ty.Type]:
    '''
    This function creates the instances of a class (ChosenClass),
    based on a scenario file name where the instances and their properties
    are given.
    '''
    scenario_vehicle: str = scenario.vehicle.name
    class_name: str = ChosenClass.class_name

    class_instances: ty.List[str] = scenario[class_name]

    instances: ty.List[ty.Type] = []

    for class_instance in class_instances:
        append_instance: bool = True
        if class_name == 'trips':
            trip_vehicle: str = scenario.trips[class_instance].vehicle
            if trip_vehicle != scenario_vehicle:
                append_instance = False
        elif class_name == 'locations':
            location_vehicle: str = scenario.locations[class_instance].vehicle
            if location_vehicle != scenario_vehicle:
                append_instance = False
        elif class_name == 'legs':
            leg_vehicle: str = scenario.legs[class_instance].vehicle
            if leg_vehicle != scenario_vehicle:
                append_instance = False

        if append_instance:
            instances.append(
                ChosenClass(class_instance, scenario, general_parameters)
            )

    return instances


# @cook.function_timer
def declare_all_instances(
    scenario: Box, case_name: str, general_parameters: Box
) -> ty.Tuple[
    pd.DataFrame, ty.List[ty.Type], ty.List[ty.Type], ty.List[ty.Type]
]:
    '''
    This declares all instances of the various objects
    (legs, locations,  trips).
    '''
    scenario_name: str = scenario.name
    file_parameters: Box = general_parameters.files
    output_folder: str = f'{file_parameters.output_root}/{case_name}'
    locations: ty.List[ty.Type] = declare_class_instances(
        Location, scenario, general_parameters
    )

    legs: ty.List[ty.Type] = declare_class_instances(
        Leg, scenario, general_parameters
    )

    # We want to get the location connections
    location_connections_headers: ty.List[str] = (
        scenario.mobility_module.location_connections_headers
    )
    location_connections_index_tuples: ty.List[ty.Tuple[str, str]] = [
        (start_location.name, end_location.name)
        for start_location in locations
        for end_location in locations
    ]
    location_connections_index: pd.MultiIndex = pd.MultiIndex.from_tuples(
        location_connections_index_tuples, names=['From', 'To']
    )
    location_connections: pd.DataFrame = pd.DataFrame(
        columns=location_connections_headers, index=location_connections_index
    )
    road_type_weights: np.ndarray = np.array(
        scenario.transport_factors.weights
    )
    for leg in legs:
        road_type_factor: float = sum(leg.road_type_mix * road_type_weights)
        fill_values: ty.List[float] = [
            leg.duration,
            leg.distance,
            road_type_factor * leg.distance,
        ]
        for fill_value, location_connection_header in zip(
            fill_values, location_connections_headers
        ):
            location_connections.loc[
                (leg.start_location, leg.end_location),
                location_connection_header,
            ] = fill_value
    pickle_interim_files: bool = general_parameters.interim_files.pickle
    if pickle_interim_files:
        location_connections.to_pickle(
            f'{output_folder}/{scenario_name}_location_connections.pkl'
        )

    trips: ty.List[ty.Type] = declare_class_instances(
        Trip, scenario, general_parameters
    )

    # We want to save the moblity matrixes
    if pickle_interim_files:
        for trip in trips:

            mobility_table_name: str = (
                f'{scenario_name}_{trip.name}_mobility_matrix'
            )

            trip.mobility_matrix.to_pickle(
                f'{output_folder}/{mobility_table_name}.pkl'
            )

            run_mobility_table_name: str = (
                f'{scenario_name}_{trip.name}_run_mobility_matrix'
            )
            trip.run_mobility_matrix.to_pickle(
                f'{output_folder}/{run_mobility_table_name}.pkl'
            )
            trip.next_leg_kilometers.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                'next_leg_kilometers.pkl'
            )
            trip.run_next_leg_kilometers.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_next_leg_kilometers.pkl'
            )
            trip.next_leg_kilometers_cumulative.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}'
                f'_next_leg_kilometers_cumulative.pkl'
            )
            trip.run_next_leg_kilometers_cumulative.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_next_leg_kilometers_cumulative.pkl'
            )
            trip.next_leg_charge_to_vehicle.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_next_leg_charge_to_vehicle.pkl'
            )
            trip.next_leg_charge_from_network.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_next_leg_charge_from_network.pkl'
            )
            trip.run_next_leg_charge_to_vehicle.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_next_leg_charge_to_vehicle.pkl'
            )
            trip.run_next_leg_charge_from_network.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_next_leg_charge_from_network.pkl'
            )
            trip.next_leg_charge_to_vehicle_cumulative.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_next_leg_charge_to_vehicle_cumulative.pkl'
            )
            trip.next_leg_charge_from_network_cumulative.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}'
                f'_next_leg_charge_from_network_cumulative.pkl'
            )
            trip.run_next_leg_charge_to_vehicle_cumulative.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}'
                f'_run_next_leg_charge_to_vehicle_cumulative.pkl'
            )
            trip.run_next_leg_charge_from_network_cumulative.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}'
                f'_run_next_leg_charge_from_network_cumulative.pkl'
            )
            trip.location_split.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_location_split.pkl'
            )
            trip.run_location_split.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_location_split.pkl'
            )
            trip.percentage_driving.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_percentage_driving.pkl'
            )
            trip.run_percentage_driving.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_percentage_driving.pkl'
            )

            trip.connectivity_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_connectivity_per_location.pkl'
            )

            trip.run_connectivity_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_connectivity_per_location.pkl'
            )

            trip.connectivity.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_connectivity.pkl'
            )

            trip.run_connectivity.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_connectivity.pkl'
            )

            trip.maximal_delivered_power_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_maximal_delivered_power_per_location.pkl'
            )
            trip.run_maximal_delivered_power_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_maximal_delivered_power_per_location.pkl'
            )

            trip.maximal_delivered_power.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_maximal_delivered_power.pkl'
            )
            trip.run_maximal_delivered_power.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_maximal_delivered_power.pkl'
            )

            trip.maximal_received_power_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_maximal_received_power_per_location.pkl'
            )
            trip.run_maximal_received_power_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_maximal_received_power_per_location.pkl'
            )

            trip.maximal_received_power.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_maximal_received_power.pkl'
            )
            trip.run_maximal_received_power.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_maximal_received_power.pkl'
            )

            trip.vehicle_discharge_power_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_vehicle_discharge_power_per_location.pkl'
            )
            trip.run_vehicle_discharge_power_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_vehicle_discharge_power_per_location.pkl'
            )

            trip.vehicle_discharge_power.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_vehicle_discharge_power.pkl'
            )
            trip.run_vehicle_discharge_power.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_vehicle_discharge_power.pkl'
            )

            trip.discharge_power_to_network_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_discharge_power_to_network_per_location.pkl'
            )
            trip.run_discharge_power_to_network_per_location.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_discharge_power_to_network_per_location.pkl'
            )

            trip.discharge_power_to_network.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_discharge_power_to_network.pkl'
            )
            trip.run_discharge_power_to_network.to_pickle(
                f'{output_folder}/{scenario_name}_'
                f'{trip.name}_run_discharge_power_to_network.pkl'
            )

            trip.battery_space_shifts_departures.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_departures'
                f'.pkl'
            )

            trip.battery_space_shifts_departures_impact.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_departures_impact'
                f'.pkl'
            )

            trip.battery_space_shifts_arrivals.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_arrivals'
                f'.pkl'
            )

            trip.battery_space_shifts_arrivals_impact.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_arrivals_impact'
                f'.pkl'
            )

            trip.battery_space_shifts_departures_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_departures_weighted'
                f'.pkl'
            )

            trip.battery_space_shifts_departures_impact_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_departures_impact_weighted'
                f'.pkl'
            )

            trip.battery_space_shifts_arrivals_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_arrivals_weighted'
                f'.pkl'
            )

            trip.battery_space_shifts_arrivals_impact_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'battery_space_shifts_arrivals_impact_weighted'
                f'.pkl'
            )

            trip.run_battery_space_shifts_departures.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_departures'
                f'.pkl'
            )

            trip.run_battery_space_shifts_departures_impact.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_departures_impact'
                f'.pkl'
            )

            trip.run_battery_space_shifts_arrivals.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_arrivals'
                f'.pkl'
            )

            trip.run_battery_space_shifts_arrivals_impact.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_arrivals_impact'
                f'.pkl'
            )

            trip.run_battery_space_shifts_departures_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_departures_weighted'
                f'.pkl'
            )

            trip.run_battery_space_shifts_departures_impact_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_departures_impact_weighted'
                f'.pkl'
            )

            trip.run_battery_space_shifts_arrivals_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_arrivals_weighted'
                f'.pkl'
            )

            trip.run_battery_space_shifts_arrivals_impact_weighted.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'run_battery_space_shifts_arrivals_impact_weighted'
                f'.pkl'
            )
            charging_sessions_headers: ty.List[str] = (
                scenario.charging_sessions.dataframe_headers
            )

            charging_sessions_dataframe: pd.DataFrame = (
                get_charging_sessions_dataframe(
                    trip.charging_sessions, scenario, charging_sessions_headers
                )
            )

            charging_sessions_dataframe.to_pickle(
                f'{output_folder}/{scenario_name}_{trip.name}_'
                f'charging_sessions'
                f'.pkl'
            )

    return location_connections, legs, locations, trips


def get_charging_sessions_dataframe(
    charging_sessions: ty.List, scenario: Box, chosen_headers: ty.List[str]
) -> pd.DataFrame:
    charging_sessions_dataframe: pd.DataFrame = pd.DataFrame(
        index=range(len(charging_sessions))
    )

    charging_sessions_properties: ty.List[str] = (
        scenario.charging_sessions.properties
    )
    for session_index, session in enumerate(charging_sessions):
        for charging_session_header, charging_session_property in zip(
            chosen_headers, charging_sessions_properties
        ):

            charging_sessions_dataframe.loc[
                session_index, charging_session_header
            ] = getattr(session, charging_session_property)

    return charging_sessions_dataframe


if __name__ == '__main__':
    start_: datetime.datetime = datetime.datetime.now()
    general_parameters_file_name: str = 'ChaProEV.toml'
    general_parameters: Box = Box(
        cook.parameters_from_TOML(general_parameters_file_name)
    )
    case_name = 'Mopo'
    test_scenario_name: str = 'XX_bus'
    scenario_file_name: str = (
        f'scenarios/{case_name}/{test_scenario_name}.toml'
    )
    scenario: Box = Box(cook.parameters_from_TOML(scenario_file_name))
    scenario.name = test_scenario_name
    location_connecions, legs, locations, trips = declare_all_instances(
        scenario, case_name, general_parameters
    )

    print((datetime.datetime.now() - start_).total_seconds())
